#!/usr/bin/env python3
"""
Script that setup users directly from swf file and launches the simulation.

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2023
"""

import os  # to issue os commands
import shutil  # to copy files
import subprocess  # to launch processes
import stat  # to change permission on a file
import time  # to wait for slurmdbd to start
import pyswf.pyswf as pyswf  # for reading swf files
import io  # to print sim_mgr stdout
import re  # to make sed-like changed
import argparse  # super comfortable way of parsing command line options


def copytree(src, dst, symlinks=False, ignore=None):
    """
    Function to copy entire directories

    taken from: https://stackoverflow.com/questions/1868714/how-do-i-copy-an-entire-directory-of-files-into-an-existing-directory-using-pyth
    :param src: source path
    :param dst: destination path
    :param symlinks: copy files to where symlinks point to (default) or copy only the pointers (might result in broken
    pointers)
    :param ignore: a callable object that encapsulates those files not to be copied
    """
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.exists(d):
            shutil.rmtree(d)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def parse_args():
    """
    Parse arguments given by the command line
    """
    parser = argparse.ArgumentParser(description='Run the BSC simulator with slurm account support',
                                     epilog="This software was adapted from https://github.com/BSC-RM/slurm_simulator "
                                            "and https://github.com/ubccr-slurm-simulator/slurm_simulator")
    parser.add_argument('swf_input', type=str, help='path where the swf file workload trace is located')
    parser.add_argument('--end-time', '-e', type=int, help='maximum simulation time in seconds')
    parser.add_argument('--looptime', '-t', type=int, help='specify the minimum real time in us that is equivalent to '
                                                           'an step increase in the simulation time. A larger looptime '
                                                           'makes the simulation slower, a smaller one, faster. Minimum'
                                                           ' value is 1.')
    parser.add_argument('--not-create-users', '-nu', dest='add_users', action='store_false', default=True,
                        help='do not add users at initialization')
    parser.add_argument('--clear-users', '-cu', action='store_true', default=False,
                        help='do not add users at initialization')

    arguments = parser.parse_args()

    # check the input
    if arguments.looptime is not None:
        if arguments.looptime < 1:
            print(f"[WARNING] Ignoring invalid looptime flag {arguments.looptime}")
            arguments.looptime = None

    if arguments.end_time is not None:
        if arguments.end_time <= 0:
            print(f"[WARNING] Ignoring invalid end time flag {arguments.end_time}")
            arguments.end_time = None

    return arguments


if __name__ == "__main__":
    args = parse_args()

    add_users = args.add_users # Adds users from the trace
    clear_users = args.clear_users  # removes all users and accounts from slurmdb

    # read the swf file
    swf_trace = pyswf.SWFtrace.from_swf(args.swf_input)
    # fetch users in the swf file
    users = swf_trace.unique_users()
    workload_name = swf_trace.input_file.split('/')[-1].split('.')[0]
    slurm_trace = '/workloads/' + swf_trace.input_file.split('/')[-1].split('.')[0] + '.trace'
    nnodes = str(swf_trace.max_nodes)
    cpus_per_node = str(int(swf_trace.max_procs/int(nnodes)))
    
    # get current directory
    cwd = os.getcwd()

    # check if the file with users exists, if not, add it
    if not os.path.exists('install/slurm_conf/users.sim'):
        with open('install/slurm_conf/users.sim', 'w') as user_file:
            user_file.write('root:0:0\n')
            user_file.write(f'{os.environ["USERNAME"]}:{os.getuid()}:{os.getgid()}\n')

    # spawn slurmdbd!
    slurmdbd = subprocess.Popen([cwd+'/install/slurm_programs/sbin/slurmdbd', "-Dvv"])

    time.sleep(2)  # wait for 2 seconds to slurmdbd to spool up, otherwise we cannot communicate with it

    if clear_users:
        # first clear users, then accounts
        sacctmgr_list_users = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-n', '-P', 'list', 'users'],
                                    shell=False,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = sacctmgr_list_users.communicate()

        for line in stdout.decode('utf-8').split("\n"):
            try:
                user, account, is_admin = line.split("|")
            except ValueError:
                continue
            if user != "root":
                sacctmgr_remove_user = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-i',
                                                         'remove', 'user', 'where', f'user={user}'], shell=False,
                                                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if sacctmgr_remove_user.wait() == 0:
                    print(f"run_simulation: Remove user {user} with association {account}")

        # clearing accounts
        sacctmgr_list_accounts = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-n', '-P', 'list',
                                                   'accounts'], shell=False, stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE)

        stdout, stderr = sacctmgr_list_accounts.communicate()

        for line in stdout.decode('utf-8').split("\n"):
            try:
                account, descr, org = line.split("|")
            except ValueError:
                continue

            if account != "root":
                sacctmgr_remove_account = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-i',
                                                     'remove', 'account', 'where', f'account={account}'], shell=False,
                                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if sacctmgr_remove_account.wait() == 0:
                    print(f"run_simulation: Remove account {account}")

        # reset users.sim file
        with open('install/slurm_conf/users.sim', 'w') as user_file:
            user_file.write('root:0:0\n')
            user_file.write(f'{os.environ["USERNAME"]}:{os.getuid()}:{os.getgid()}\n')

    if add_users:  # if flag is true and there is a trace to read from

        consumed_gids = []  # track those gids that where added

        # add users
        with open('install/slurm_conf/users.sim', 'a') as user_file:
            for user in users:
                uid, gid = user
                username = "u" + str(uid)
                account = "g" + str(gid)
                fairshare = 100  # TODO MAKE THIS VARIABLE?
                if gid not in consumed_gids:
                    # launch sacctmgr to create account if not already processed
                    # -i is to make immediate the changes
                    # fairshare of the account
                    sacctmgr_account = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-i', 'add',
                                                     'account', f'account={account}', f'fairshare={fairshare}'],
                                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    if sacctmgr_account.wait() == 0:
                        print(f"run_simulation: Add account {account} with fairshare {fairshare}")
                    consumed_gids.append(gid)

                # launch sacctmgr to create user uid under account g+gid
                sacctmgr_user = subprocess.Popen([cwd + '/install/slurm_programs/bin/sacctmgr', '-i', 'add', 'user',
                                            f'user={username}', f'DefaultAccount={account}'], stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE)

                if sacctmgr_user.wait() == 0:
                    print(f"run_simulation: Add user {user} to account {account}")
                user_file.write(f'{username}:{uid}:{gid}\n')

    # convert to bsc trace, then to binary trace
    # bsc's swf file format
    bsc_trace = '/workloads/' + swf_trace.input_file.split('/')[-1].split('.')[0] + '.bscswf'

    # convert the swf to bsc's swf
    if not os.path.isfile(cwd + bsc_trace):
        swf_trace.to_bsc_swf(cwd + bsc_trace, int(swf_trace.max_jobs))
        print(f"run_simulation: Convert swf file to bscswf")

    # convert bsc's swf to binary trace, if not already present
    if not os.path.isfile(cwd + slurm_trace):
        swf2trace = subprocess.Popen([cwd + '/ces_slurm_simulator/contribs/simulator/swf2trace', cwd + bsc_trace,
                                      cwd + slurm_trace, cpus_per_node, str(swf_trace.max_jobs)],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for out_line in iter(swf2trace.stdout):
            print(out_line)
        if swf2trace.wait() == 0:
            print(f"run_simulation: Convert bscswf file to binary trace")

    # this is translation of run_simulation_local.sh

    sim_path = cwd + "/simulation"
    user = os.environ["USERNAME"]

    # copy install contents to the new folder
    os.mkdir(sim_path + "/slurm_varios")
    os.mkdir(sim_path + "/slurm_programs")
    copytree(cwd + "/install/slurm_varios/", sim_path + "/slurm_varios/")
    copytree(cwd + "/install/slurm_programs/", sim_path + "/slurm_programs/")
    shutil.copyfile(cwd + "/install/slurm_conf/slurm.conf.template", sim_path + "/slurm_conf/")

    slurm_conf_template = f"{sim_path}/slurm_conf/slurm.conf.template"
    slurmctld_port = os.getpid() % 65535  # 12 is the max number of jobs that can enter mn4 node
    port_range = os.getpid() % 12
    slurmctld_port = (slurmctld_port * port_range) % 65535

    if slurmctld_port <= 10000:
        slurmctld_port += 10000

    slurmctld_f_port = slurmctld_port + 12
    slurmd_port = slurmctld_port+144

    # generate rsa key and certificate for job submission
    openssl = subprocess.Popen(['openssl', 'genrsa', f'{sim_path}/slurm_conf/slurm.key', '1024'],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    openssl.wait()
    openssl = subprocess.Popen(['openssl', 'rsa', '-in', f'{sim_path}/slurm_conf/slurm.key', '-pubout',
                                f'{sim_path}/slurm_conf/slurm.cert'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    openssl.wait()

    # adapt job trace script to this simulation
    with open(f'{sim_path}/slurm_varios/trace.sh.template', 'r') as trace_template:
        trace_lines = trace_template.readlines()
        with open(f"{sim_path}/slurm_varios/trace.sh", "w") as trace_script:
            for line in trace_lines:
                re_line = re.sub("{DIR_WORK}", sim_path, line)
                trace_script.write(re_line)

    # allow the execution of the script
    st = os.stat(f"{sim_path}/slurm_varios/trace.sh")

    # this script is passed to slurm, and it is the one that controls the output of sim_mgr
    os.chmod(f"{sim_path}/slurm_varios/trace.sh", st.st_mode | stat.S_IEXEC)

    # create slurm configuration file
    with open(slurm_conf_template, "r") as conf_template:
        conf_lines = conf_template.readlines()
        with open(f"{sim_path}/slurm_conf/slurm.conf", "w") as slurm_conf:
            for line in conf_lines:
                re_line = re.sub("TOKEN_BF_QUEUE", "1000", line)
                re_line = re.sub("TOKEN_CONTROL_MACHINE", "localhost", re_line)
                re_line = re.sub("TOKEN_SLURMCTLD_PORT", f"{slurmctld_port}-{slurmctld_f_port}", re_line)
                re_line = re.sub("TOKEN_SLURMD_PORT", str(slurmd_port), re_line)
                re_line = re.sub("TOKEN_SLURM_USER_PATH", sim_path, re_line)
                re_line = re.sub("TOKEN_USER", user, re_line)
                re_line = re.sub("TOKEN_NNODES", nnodes, re_line)
                re_line = re.sub("TOKEN_NCORESPERNODE", cpus_per_node, re_line)
                slurm_conf.write(re_line)

    # run source command
    # set up environment variables
    os.environ["SIM_DAEMONS_PATH"] = f"{sim_path}/slurm_programs/sbin"
    os.environ["SIM_MGR_PATH"] = f"{sim_path}/slurm_programs/bin"
    os.environ["SIM_DIR"] = f"{sim_path}"
    # append sim path's to PATH variable
    os.environ["PATH"] += os.pathsep + os.pathsep.join([os.environ["SIM_DAEMONS_PATH"], os.environ["SIM_MGR_PATH"],
                                                        os.environ["SIM_DIR"]])
    os.environ["SLURM_CONF"] = f'{sim_path}/slurm_conf/slurm.conf'
    os.environ["SLURM_SIM_ID"] = str(os.getpid())

    # reset user's and account's rawusage if they were not recently added
    if not add_users:
        consumed_gids = []

        # launch slurmctld
        slurmctld = subprocess.Popen(['slurmctld', '-c', '-D'], stdout=subprocess.PIPE)
        for user in users:
            uid, gid = user
            username = "u"+str(uid)
            account = "g"+str(gid)
            sacctmgr_user = subprocess.Popen(['sacctmgr', '-i', 'modify', 'user', 'where'
                                          f'user={username}', f'account={account}', 'set',
                                          f'rawusage=0'], stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
            if gid not in consumed_gids:
                sacctmgr_user = subprocess.Popen(
                    ['sacctmgr', '-i', 'modify', 'account', 'where', f'account={account}', 'set', f'rawusage=0'],
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                consumed_gids.append(gid)

        # FINISH slurmctld, to allow sim_mgr to launch his own
        slurmctld.kill()

    # fork and point to the binary workload file
    command = ['sim_mgr', '-f', '-w', cwd + slurm_trace]

    if args.looptime is not None:
        command += ['-t', f'{args.looptime}']

    if args.end_time is not None:
        command.append(f'{args.end_time}')

    with subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as sim_mgr, \
            io.StringIO() as buf:
        for line in sim_mgr.stdout:
            print(line, end="")
            buf.write(line)

    sim_mgr.kill()

    # finish slurm dbd
    slurmdbd.kill()
